@@ -165,6 +165,10 @@ func New(
 	va *va.VAImpl,
 	ca *ca.CAImpl,
 	strict, requireEAB bool) WebFrontEndImpl {
+	// Seed rand from the current time so test environments don't always have
+	// the same nonce rejection and sleep time patterns.
+	rand.Seed(time.Now().UnixNano())
+
 	// Read the % of good nonces that should be rejected as bad nonces from the
 	// environment
 	nonceErrPercentVal := os.Getenv(badNonceEnvVar)
@@ -245,6 +249,11 @@ func (wfe *WebFrontEndImpl) HandleFunc(
 	defaultHandler := http.StripPrefix(pattern,
 		&topHandler{
 			wfe: wfeHandlerFunc(func(ctx context.Context, response http.ResponseWriter, request *http.Request) {
+				// Process CORS as necessary. If it's a CORS preflight, no further processing may occur.
+				if wfe.processCORS(request, response, methodsMap) {
+					return
+				}
+
 				// Modern ACME only sends a Replay-Nonce in responses to GET/HEAD
 				// requests to the dedicated newNonce endpoint, or in replies to POST
 				// requests that consumed a nonce.
@@ -280,6 +289,46 @@ func (wfe *WebFrontEndImpl) HandleFunc(
 	mux.Handle(pattern, defaultHandler)
 }
 
+// processCORS reads and writes all necessary request and response headers in order to
+// enable use by CORS-aware user agents. If the request is a CORS preflight request, the
+// function returns true, in which case no further data may be written to the response.
+func (wfe *WebFrontEndImpl) processCORS(request *http.Request, response http.ResponseWriter,
+	allowedMethodsMap map[string]bool) bool {
+	// 6.1.1, 6.2.1. No Origin header means CORS is not relevant.
+	// 6.1.2, 6.2.2. Origin's value is not processed because it always matches.
+	if request.Header.Get("Origin") == "" {
+		return false
+	}
+
+	// 6.2. Request is a CORS preflight
+	if request.Method == http.MethodOptions {
+		// 6.2.3, 6.2.5. -Request-Method must be present and must be a match for one of the allowed methods
+		method := request.Header.Get("Access-Control-Request-Method")
+		if _, allowed := allowedMethodsMap[method]; method == "" || !allowed {
+			return false
+		}
+		// 6.2.4, 6.2.6. -Request-Headers i not processed because it always matches.
+		// 6.2.7. Send -Allow-Origin, without -Allow-Credentials support.
+		response.Header().Set("Access-Control-Allow-Origin", "*")
+		// 6.2.8. Send -Max-Age.
+		response.Header().Set("Access-Control-Max-Age", "5")
+		// 6.2.9. -Allow-Methods is not sent because ACME only uses simple methods.
+		// 6.2.10. -Allow-Headers required for "Content-Type: application/jose+json"
+		response.Header().Set("Access-Control-Allow-Headers", "Content-Type")
+		// Terminate the request
+		response.WriteHeader(http.StatusNoContent)
+		return true
+	}
+
+	// 6.1. Otherwise, request is a CORS simple or actual request.
+	// 6.1.3. Send -Allow-Origin, without Allow-Credentials support.
+	response.Header().Set("Access-Control-Allow-Origin", "*")
+	// 6.1.4. Send -Expose-Headers for the response headers ACME uses.
+	response.Header().Set("Access-Control-Expose-Headers", "Link, Replay-Nonce, Location")
+	// Continue processing the request
+	return false
+}
+
 func (wfe *WebFrontEndImpl) HandleManagementFunc(
 	mux *http.ServeMux,
 	pattern string,
@@ -2353,14 +2402,14 @@ func (wfe *WebFrontEndImpl) Certificate(
 		return
 	}
 
-	if no >= len(cert.Issuers) {
+	if no >= len(cert.IssuerChains) {
 		response.WriteHeader(http.StatusNotFound)
 		return
 	}
 
 	// Add links to alternate roots
 	basePath := wfe.relativeEndpoint(request, fmt.Sprintf("%s%s", certPath, serial))
-	addAlternateLinks(response, basePath, no, len(cert.Issuers))
+	addAlternateLinks(response, basePath, no, len(cert.IssuerChains))
 
 	response.Header().Set("Content-Type", "application/pem-certificate-chain; charset=utf-8")
 	response.WriteHeader(http.StatusOK)
